
// Subscripts: ranges and embeddings

#include <stdlib.h>
#include "Seq.h"
#include "Procs.h"
#include "Range.h"
extern "C" {
  int printf(char*, ...) ;
}  // debug
extern "C" {
  #include "mpi.h"
}  // debug


// [This implementation of `Range' fudges mapping of multiple virtual
// processes to individual physical processes by scaling up the range
// blocksize, then always dealing directly with physical processes.
//
// That doesn't always *strictly* implement distribution formats of
// HPF, when there are more virtual than physical processes.  It is
// a reasonable approximation, however.]

// Lines labelled ``bodge'' are clumsy fixes to support the Fortran
// run-time library.


// Constructors for primitive ranges

Range :: Range(int _extent) : Subscript(1) {
  extent    = _extent ;
  dim       = NULL    ;
  blockSize = extent  ;
  layerSize = blockSize ;
  base      = 0 ;
  stride    = 1 ;
}

Range :: Range(int _extent, Dimension* _dim, Format fmt) : Subscript(1) {
  extent    = _extent ;
  dim       = _dim  ;
  switch(fmt) {
    case blk :
      blockSize = (extent + dim->extent - 1) / dim->extent ;
      break ;
    case cyc :
      blockSize = dim->blockSize ;
      break ;
  }
  layerSize = dim->extent * blockSize ;
  base      = 0 ;
  stride    = 1 ;
}

Range :: Range(int _extent, Dimension* _dim, Format fmt, int _blockSize) :
            Subscript(1) {
  extent    = _extent ;
  dim       = _dim  ;
  blockSize = dim->blockSize * _blockSize ;
  layerSize = dim->extent * blockSize ;
  base      = 0 ;
  stride    = 1 ;
  if(fmt == blk & extent > layerSize) {
    //printf("Range :: Range: dim->blockSize = %d _blockSize = %d dim->extent = %d\n", dim->blockSize, _blockSize, dim->extent) ;  // debug
    //printf("Range :: Range: extent = %d layerSize = %d\n", extent, layerSize) ;  // debug
    AdlibError("Range :: Range: extent too large for specified block size") ;
  }
}


// Constructor for derived ranges

Range :: Range(int _extent, int _base, int _stride, Range& parent) :
             Subscript(1) {
  extent    = _extent ;
  dim       = parent.dim ;
  blockSize = parent.blockSize ;
  layerSize = parent.layerSize ;
  base      = parent.base + parent.stride * _base ;
  stride    = parent.stride * _stride ;
}

// Iterator functions

//    s_v, s_h: Horizontal and vertical coordinates in local portion of
//              parent primitive range.
//
//           i: Range index.


void Range :: blockRange(int* i_lo, int* i_hi) {

  // Define `i_lo' and `i_hi' for current values of `s_v', `coord'

  // No guarantees are made about the results if `s_v' and `coord'
  // do not define a position within the bounds of the range.

  const int blockStart = s_v * layerSize + coord * blockSize ;

  const int blockStartOffset  = blockStart - base ;
  const int blockFinishOffset = blockStartOffset + blockSize - 1 ;

  if(stride > 0) {
    *i_lo = max((blockStartOffset + stride - 1) / stride, 0) ;
    *i_hi = min(blockFinishOffset / stride, extent - 1) ;
  }
  else {
    *i_lo = max((blockFinishOffset + stride + 1) / stride, 0) ;
    *i_hi = min(blockStartOffset / stride, extent - 1) ;
  }
}


// Inquiry functions for use during block iteration

int Range :: blockBase()  {return i ; } 
int Range :: blockCount() {return max(i_limit - i + 1, 0) ; }
int Range :: blockCoord() {return coord ; }


// Iteration over locally held blocks.

void Range :: locBlocksEnum() {
  s_v = base / layerSize ;

  if(dim) {
    const int baseCoord = (base / blockSize) % dim->extent ;

    coord = dim->coord ;
    if(stride > 0 && coord < baseCoord) s_v++ ;
    if(stride < 0 && coord > baseCoord) s_v-- ;
  }
  else
    coord = 0 ;

  blockRange(&i, &i_limit) ;
  bodge = i ;  // bodge
  s_h = (base + stride * i) % blockSize ;

  if(i > i_limit && locBlocksTest())
    locBlocksNext() ;
}

int Range :: locBlocksTest() { return i < extent ; }

void Range :: locBlocksNext() {
  do { 
    if(stride > 0)
      s_v++ ;
    else
      s_v-- ;

    blockRange(&i, &i_limit) ;
    bodge = i ;  // bodge
    s_h = (base + stride * i) % blockSize ;

  } while (i > i_limit && locBlocksTest()) ;
}


// Iteration over all blocks of range.

void Range :: allBlocksEnum() {
  s_v   = base / layerSize ;
  coord = dim ? (base / blockSize) % dim->extent : 0 ;

  blockRange(&i, &i_limit) ;
  bodge = i ;  // bodge
  s_h = (base + stride * i) % blockSize ;

  if(i > i_limit && allBlocksTest())
    allBlocksNext() ;
}

int Range :: allBlocksTest() { return i < extent ; }

void Range :: allBlocksNext() {
  do {
    if(stride > 0)
      if(dim) {
	coord++ ;
	if(coord == dim->extent) {
	  coord = 0 ;
	  s_v++ ;
	}
      }
      else
	s_v++ ;
    else
      if(dim) {
	coord-- ;
	if(coord == -1) {
	  coord += dim->extent ;
	  s_v-- ;
	}
      }
      else
	s_v-- ;

    blockRange(&i, &i_limit) ;
    bodge = i ;  // bodge
    s_h = (base + stride * i) % blockSize ;

  } while (i > i_limit && allBlocksTest()) ;
}


// Iterators for `where' loop.

void Range :: where() {
  isRange = 0 ;

  if(dim)
    apg->deactivate(this) ;

  locBlocksEnum() ;
}

Range :: operator void*() {
  return (void*) locBlocksTest() ;
}

void Range :: operator++() {
  i++ ;
  bodge = i ;  // bodge
  s_h += stride ;

  if(i > i_limit && locBlocksTest())
    locBlocksNext() ;
}

void Range :: erewh() {
  if(dim)
    apg->reactivate(this) ;

  isRange = 1 ;
}


// Subscripting of Range.

int Range :: operator()(Subscript& i) {
  const int primIdxVal = layerSize * i.s_v + blockSize * coord + i.s_h ;

  return (primIdxVal - base) / stride ;
}

int Range :: operator()() {
  return i ;
}


// Constructors for embeddings.

Embed :: Embed(int _location, Range& x) : Subscript(0) {
  dim       = x.dim ;
  blockSize = x.blockSize ;
  layerSize = x.layerSize ;

  location  = x.base + x.stride * _location ;
  coord = (location % layerSize) / blockSize ;
  s_v =  location / layerSize ;
  s_h =  location % blockSize ;

  bodge = _location ;  // bodge
}

//// Dummy copy constructor (no foreseen need to copy an embedding).
//
//Embed :: Embed(const Embed& e) : Subscript(0) {
//  AdlibError("Embed :: Embed: copying of embeddings not supported") ;
//}


// Test if embedding is held locally.

int Embed :: at() {
  if(dim)
    return apg->deactivate(this) ;
  else
    return 1 ;
}

void Embed :: ta() {
  if(dim)
    apg->reactivate(this) ;
}

// ---------------------------------------------------
//
// Things to do:
//
//   Consider putting the current functionality of
//     int Range :: operator()() 
//   into
//     Range :: operator int() 
//   This may involve change the interface of
//     Range :: operator void*() 
//   to avoid confusion.
//
//   Range :: Range(int, Dimension*, Format)
//     probably it should work in terms of virtual extents, so that
//     the blocksize is a multiple of what strict HPF implies?
//
//   Rationalise constructors for `Subscript'.
//
//   Change `blk' and `cyc' to `BLK' and `CYC'.
//
//   Introduce a `Triplet' class?  Provide coercions to enable their
//   use (and use of integers) as subscripts.
//
//   Introduce more intelligent allocation strategy for `Range' structures.
//   Really need allocation from a pool, with reference-counting.
//   As things work at the moment, the only *safe* way of creating
//   range objects is to declare named ones.

